# autogenerated by __main__.py do not edit
name_map = {
    'QDBus': 'PyQt6.QtDBus',
    'QDBusAbstractAdaptor': 'PyQt6.QtDBus',
    'QDBusAbstractInterface': 'PyQt6.QtDBus',
    'QDBusArgument': 'PyQt6.QtDBus',
    'QDBusConnection': 'PyQt6.QtDBus',
    'QDBusConnectionInterface': 'PyQt6.QtDBus',
    'QDBusError': 'PyQt6.QtDBus',
    'QDBusInterface': 'PyQt6.QtDBus',
    'QDBusMessage': 'PyQt6.QtDBus',
    'QDBusObjectPath': 'PyQt6.QtDBus',
    'QDBusPendingCall': 'PyQt6.QtDBus',
    'QDBusPendingCallWatcher': 'PyQt6.QtDBus',
    'QDBusPendingReply': 'PyQt6.QtDBus',
    'QDBusReply': 'PyQt6.QtDBus',
    'QDBusServiceWatcher': 'PyQt6.QtDBus',
    'QDBusSignature': 'PyQt6.QtDBus',
    'QDBusUnixFileDescriptor': 'PyQt6.QtDBus',
    'QDBusVariant': 'PyQt6.QtDBus',
    'QtDBus': 'PyQt6.QtDBus',
}
module_names = frozenset(('QtDBus',))
