#!/usr/bin/env python
# License: GPLv3 Copyright: 2008, Kovid Goyal kovid@kovidgoyal.net

"""
Convert CBR/CBZ files to LRF.
"""

import sys


def main(args=sys.argv):
    return 0


if __name__ == '__main__':
    sys.exit(main())
