# License: GPLv3 Copyright: 2010, Hiroshi Miura <miurahr@linux.com>

"""
Decode unicode text to an ASCII representation of the text in Vietnamese.

"""

from calibre.ebooks.unihandecode.unicodepoints import CODEPOINTS
from calibre.ebooks.unihandecode.unidecoder import Unidecoder
from calibre.ebooks.unihandecode.vncodepoints import CODEPOINTS as HANCODES


class Vndecoder(Unidecoder):
    codepoints = {}

    def __init__(self):
        self.codepoints = CODEPOINTS
        self.codepoints.update(HANCODES)
