#!/usr/bin/env python
# License: GPLv3 Copyright: 2015, Kovid Goyal <kovid at kovidgoyal.net>

EXT_MAP = {
    'default': 'unknown',
    'dir': 'dir',
    'zero': 'zero',
    'jpeg': 'jpeg',
    'jpg': 'jpeg',
    'gif': 'gif',
    'png': 'png',
    'bmp': 'bmp',
    'cbz': 'cbz',
    'cbr': 'cbr',
    'svg': 'svg',
    'html': 'html',
    'htmlz': 'html',
    'htm': 'html',
    'xhtml': 'html',
    'xhtm': 'html',
    'lit': 'lit',
    'lrf': 'lrf',
    'lrx': 'lrx',
    'pdf': 'pdf',
    'pdr': 'zero',
    'rar': 'rar',
    'zip': 'zip',
    'txt': 'txt',
    'text': 'txt',
    'prc': 'mobi',
    'azw': 'mobi',
    'mobi': 'mobi',
    'pobi': 'mobi',
    'mbp': 'zero',
    'azw1': 'tpz',
    'azw2': 'azw2',
    'azw3': 'azw3',
    'kfx': 'kfx',
    'kfx-zip': 'kfx',
    'azw4': 'pdf',
    'tpz': 'tpz',
    'tan': 'zero',
    'epub': 'epub',
    'fb2': 'fb2',
    'rtf': 'rtf',
    'odt': 'odt',
    'snb': 'snb',
    'djv': 'djvu',
    'djvu': 'djvu',
    'xps': 'xps',
    'oxps': 'xps',
    'docx': 'docx',
    'opml': 'opml',
}
