#!/usr/bin/env python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai
# License: GPLv3 Copyright: 2011, Kovid Goyal <kovid@kovidgoyal.net>

from __future__ import absolute_import, division, print_function, unicode_literals
