# autogenerated by __main__.py do not edit
name_map = {
    'PYQT_WEBENGINE_VERSION': 'PyQt6.QtWebEngineCore',
    'PYQT_WEBENGINE_VERSION_STR': 'PyQt6.QtWebEngineCore',
    'QWebChannel': 'PyQt6.QtWebChannel',
    'QWebChannelAbstractTransport': 'PyQt6.QtWebChannel',
    'QWebEngineCertificateError': 'PyQt6.QtWebEngineCore',
    'QWebEngineClientCertificateSelection': 'PyQt6.QtWebEngineCore',
    'QWebEngineClientCertificateStore': 'PyQt6.QtWebEngineCore',
    'QWebEngineClientHints': 'PyQt6.QtWebEngineCore',
    'QWebEngineContextMenuRequest': 'PyQt6.QtWebEngineCore',
    'QWebEngineCookieStore': 'PyQt6.QtWebEngineCore',
    'QWebEngineDesktopMediaRequest': 'PyQt6.QtWebEngineCore',
    'QWebEngineDownloadRequest': 'PyQt6.QtWebEngineCore',
    'QWebEngineExtensionInfo': 'PyQt6.QtWebEngineCore',
    'QWebEngineExtensionManager': 'PyQt6.QtWebEngineCore',
    'QWebEngineFileSystemAccessRequest': 'PyQt6.QtWebEngineCore',
    'QWebEngineFindTextResult': 'PyQt6.QtWebEngineCore',
    'QWebEngineFrame': 'PyQt6.QtWebEngineCore',
    'QWebEngineFullScreenRequest': 'PyQt6.QtWebEngineCore',
    'QWebEngineGlobalSettings': 'PyQt6.QtWebEngineCore',
    'QWebEngineHistory': 'PyQt6.QtWebEngineCore',
    'QWebEngineHistoryItem': 'PyQt6.QtWebEngineCore',
    'QWebEngineHistoryModel': 'PyQt6.QtWebEngineCore',
    'QWebEngineHttpRequest': 'PyQt6.QtWebEngineCore',
    'QWebEngineLoadingInfo': 'PyQt6.QtWebEngineCore',
    'QWebEngineNavigationRequest': 'PyQt6.QtWebEngineCore',
    'QWebEngineNewWindowRequest': 'PyQt6.QtWebEngineCore',
    'QWebEngineNotification': 'PyQt6.QtWebEngineCore',
    'QWebEnginePage': 'PyQt6.QtWebEngineCore',
    'QWebEnginePermission': 'PyQt6.QtWebEngineCore',
    'QWebEngineProfile': 'PyQt6.QtWebEngineCore',
    'QWebEngineProfileBuilder': 'PyQt6.QtWebEngineCore',
    'QWebEngineQuotaRequest': 'PyQt6.QtWebEngineCore',
    'QWebEngineRegisterProtocolHandlerRequest': 'PyQt6.QtWebEngineCore',
    'QWebEngineScript': 'PyQt6.QtWebEngineCore',
    'QWebEngineScriptCollection': 'PyQt6.QtWebEngineCore',
    'QWebEngineSettings': 'PyQt6.QtWebEngineCore',
    'QWebEngineUrlRequestInfo': 'PyQt6.QtWebEngineCore',
    'QWebEngineUrlRequestInterceptor': 'PyQt6.QtWebEngineCore',
    'QWebEngineUrlRequestJob': 'PyQt6.QtWebEngineCore',
    'QWebEngineUrlScheme': 'PyQt6.QtWebEngineCore',
    'QWebEngineUrlSchemeHandler': 'PyQt6.QtWebEngineCore',
    'QWebEngineView': 'PyQt6.QtWebEngineWidgets',
    'QWebEngineWebAuthPinRequest': 'PyQt6.QtWebEngineCore',
    'QWebEngineWebAuthUxRequest': 'PyQt6.QtWebEngineCore',
    'QtWebChannel': 'PyQt6.QtWebChannel',
    'QtWebEngineCore': 'PyQt6.QtWebEngineCore',
    'QtWebEngineWidgets': 'PyQt6.QtWebEngineWidgets',
    'qWebEngineChromiumSecurityPatchVersion': 'PyQt6.QtWebEngineCore',
    'qWebEngineChromiumVersion': 'PyQt6.QtWebEngineCore',
    'qWebEngineGetDomainAndRegistry': 'PyQt6.QtWebEngineCore',
    'qWebEngineProcessName': 'PyQt6.QtWebEngineCore',
    'qWebEngineVersion': 'PyQt6.QtWebEngineCore',
}
module_names = frozenset(('QtWebEngineCore', 'QtWebEngineWidgets', 'QtWebChannel'))
